/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */


#include "environment.h"
#include "globaldata.h"
#include "explosion.h"
#include "missile.h"
#include "decor.h"
#include "tank.h"

MISSILE::MISSILE ()
{
	_align = LEFT;
	_current.x = 0;
	_current.y = 0;
	_current.w = 0;
	_current.h = 0;
}

MISSILE::~MISSILE ()
{
	_env->removeObject (this);
}

MISSILE::MISSILE (GLOBALDATA *global, ENVIRONMENT *env, double xpos, double ypos, double xvel, double yvel, int weaponType)
{
	initialise ();
	setEnvironment (env);
	player = NULL;
	_align = LEFT;
	_global = global;
	x = xpos;
	y = ypos;
	xv = xvel;
	yv = yvel;
	type = weaponType;
	if (type < global->weapons)
		weap = &weapon[type];
	else
		weap = &natural[type - global->weapons];
	mass = weap->mass;
	drag = weap->drag;
	grip = weap->grip;
	sound = weap->sound;
	expSize = weap->radius;
	etime = weap->etime;
	damage = weap->damage;
	eframes = weap->eframes;
	trigger = weap->trigger;
	physicsType = weap->physicsType;
	objRadius = weap->objRadius;
	inertia = 0.5 * mass * objRadius * objRadius;
	dispersionType = weap->dispersionType;
	disperseDirection = weap->disperseDirection;
	comeToRestPeriod = (int)(weap->comeToRestPeriod * FRAMES_PER_SECOND);
	comeToRestArea = weap->comeToRestArea;
	elasticity = weap->elasticity;
	countdown = -1;
	_bitmap = weap->inGameImage;

	if (trigger == TRIGGER_COMETOREST) {
		positionHistory = new VECTOR[comeToRestPeriod];
		for (int i = 0; i < comeToRestPeriod; i++)
			positionHistory[i].x = positionHistory[i].y = 0;
	}
}

void MISSILE::initialise ()
{
	PHYSICAL_OBJECT::initialise ();
	physics = 0;
	age = 0;
	if (type >= SML_METEOR && type <= LRG_METEOR) {
		angle = rand () % 360;
		spin = rand () % 20 - 10;
	} else {
		angle = 0;
		spin = 0;
	}
}

void MISSILE::explode ()
{
	EXPLOSION *explosion;

	explosion = new EXPLOSION (_global, _env, x, y, type);
	if (!explosion) {
		perror ("missile.cc: Failed allocating memory for explosion in MISSILE::explode");
		exit (1);
	}
	explosion->player = player;
	destroy = TRUE;
}

int MISSILE::triggerTest ()
{
	int quell = 1;
	TANK *tank;

	// Has it hit a tank?
	for (int objCount = 0; (tank = (TANK*)_env->getNextOfClass (TANK_CLASS, &objCount)) && tank; objCount++) {
		if (x > tank->x - TANKWIDTH && x < tank->x + TANKWIDTH && y > tank->y && y < tank->y + TANKHEIGHT && tank->l > 0) {
			hitSomething = 1;
			tank->requireUpdate ();
		}
	}
	
	if (trigger == TRIGGER_NONE) {
		quell = 1;
	} else if (trigger == TRIGGER_COLLISION) {
		if ((y > 0) && (hitSomething || (y >= _global->screenHeight) || (getpixel (_env->terrain, (int)x, (int)y) != PINK))) {
			quell = 0;
		}
	} else if (trigger == TRIGGER_COMETOREST) {
		int position;
		for (position = 1; position < comeToRestPeriod; position++) {
			if (fabs (x - positionHistory[position].x) +
				fabs (y - positionHistory[position].y)
				> comeToRestArea)
				break;
		}
		if (position == comeToRestPeriod)
			quell = 0;
		for (position = comeToRestPeriod - 1; position > 0; position--) {
			positionHistory[position] = positionHistory[position - 1];
		}
		positionHistory[0].x = x;
		positionHistory[0].y = y;
	} else if (trigger == TRIGGER_RESURFACE) {
		if (physics == 1) {
			if (getpixel (_env->terrain, (int)x, (int)y) == PINK)
				quell = 0;
			else
				quell = 1;
		} else if (!physics) {
			if (y < 0) {
				quell = 1;
			} else if (getpixel (_env->terrain, (int)x, (int)y) != PINK) {
				physics = 1;
				xv *= 0.1;
				yv *= 0.1;
				quell = 1;
			}
		}
	} else if (trigger == TRIGGER_ZENITH) {
		if (yv > 0) {
			quell = 0;
		}
	} else if (trigger == TRIGGER_INSTANT) {
		quell = 0;
		destroy = TRUE;
	}

	if (countdown >= 0 && age >= countdown) {
		quell = 0;
	}
	if (!quell) {
		_env->realm--;
		activate ();
	}
	

	return (!quell);
}

int MISSILE::applyPhysics ()
{
	int detonate = 0;
	age++;

	if (age > 15 * FRAMES_PER_SECOND || y < -65535)
		detonate = 1;

	PHYSICAL_OBJECT::applyPhysics ();

	if (detonate)
		hitSomething = 1;

	return (detonate);
}


void MISSILE::activate ()
{
	EXPLOSION *explosion;

	if (weap->submunition >= 0) {
		WEAPON *submunition = &weapon[weap->submunition];

		if (weap->numSubmunitions > 0) {
			double divergenceStep;
			if (weap->numSubmunitions > 1)
				divergenceStep = weap->divergence / (weap->numSubmunitions - 1);
			else
				divergenceStep = 0;
			int startPoint;
			int randStart = rand () % 1000000;
			double inheritedXV = weap->impartVelocity * xv;
			double inheritedYV = weap->impartVelocity * yv;

			VECTOR dispersalPoint = VECTOR (x, y);
			if (disperseDirection == DISPERSEDIR_DOWN) {
				startPoint = 0;
				dispersalPoint = dispersalPoint + VECTOR (0, 10);
			} else if (disperseDirection == DISPERSEDIR_UP) {
				startPoint = 180;
				dispersalPoint = dispersalPoint - VECTOR (0, 10);
			} else if (disperseDirection == DISPERSEDIR_FORWARDS) {
				//int directionOfTravel = (int)(atan2 (yv, xv) / PI * 180) + 90;
				int directionOfTravel = (int)(atan2 (yv, xv) / PI * 180);
				if (directionOfTravel < 0)
					directionOfTravel += 360;
				startPoint = directionOfTravel;
				dispersalPoint = dispersalPoint + (_global->slope[startPoint] * 20);
			}
			for (int spreadCount = 0; spreadCount < weap->numSubmunitions; spreadCount++) {
				MISSILE *newmis;
				double launchSpeed;
				int newmisCount;
				int dAngle;

				if (dispersionType == DISPERSE_SPHERICAL) {
				       	dAngle = (int) (startPoint - (weap->divergence / 2) + (divergenceStep * spreadCount));
					if (weap->spreadVariation > 0) {
						double variation = Noise (randStart + 1054 + spreadCount) * weap->spreadVariation;
						dAngle += (int)(weap->divergence * variation);
					}
				} else if (dispersionType == DISPERSE_PARALLEL) {
					dAngle = startPoint - 90;
				}
				while (dAngle < 0)
					dAngle += 360;
				while (dAngle >= 360)
					dAngle -= 360;

				newmisCount = submunition->countdown;
				if (submunition->countVariation > 0) {
					double variation = Noise (randStart + 78689 + spreadCount) * submunition->countVariation;
					newmisCount += (int)(submunition->countdown * variation);
					if (newmisCount <= 0)
						newmisCount = 1;
				}
				
				if (dispersionType == DISPERSE_SPHERICAL) {
					launchSpeed = weap->launchSpeed;
				} else if (dispersionType == DISPERSE_PARALLEL) {
					launchSpeed = weap->launchSpeed;
				       	launchSpeed = (launchSpeed) - (launchSpeed * 2 / weap->numSubmunitions * spreadCount);
				}
				if (weap->speedVariation > 0) {
					double variation = Noise (randStart + 124786 + spreadCount) * weap->speedVariation;
					launchSpeed += weap->launchSpeed * variation;
				}

				newmis = new MISSILE (_global, _env, dispersalPoint.x, dispersalPoint.y, _global->slope[dAngle].x * launchSpeed * (100.0 / FRAMES_PER_SECOND) + inheritedXV, _global->slope[dAngle].y * launchSpeed * (100.0 / FRAMES_PER_SECOND) + inheritedYV, weap->submunition);

				if (!newmis) {
					perror ("missile.cc: Failed allocating memory for newmis in MISSILE::triggerTest (CLUSTER)");
					exit (1);
				}

				newmis->physics = weapon[weap->submunition].physicsType;
				newmis->player = player;
				newmis->countdown = newmisCount;
				newmis->trigger = weapon[weap->submunition].trigger;
				newmis->setUpdateArea
					((int) newmis->x - 20,
					 (int) newmis->y - 20, 40, 40);
			}
		}
	} else {
		if (weap->detonationType == DETONATIONTYPE_EARTHQUAKE)
			explosion = new EXPLOSION (_global, _env, x, y, xv, yv, type);
		else if (weap->detonationType == DETONATIONTYPE_ERASEWEDGE)
			explosion = new EXPLOSION (_global, _env, x, y, xv, yv, type);
		else
			explosion = new EXPLOSION (_global, _env, x, y, type);

		if (!explosion) {
			perror ("missile.cc: Failed allocating memory for explosion in MISSILE::activate");
			exit (1);
		}
		explosion->player = player;
	}
	destroy = TRUE;
	
	if (weap->detonationType == DETONATIONTYPE_DIRT ||
		weap->detonationType == DETONATIONTYPE_ERASEWEDGE) {
		play_sample ((SAMPLE *) _global->SOUND[9].dat, 128 + (expSize / 2), 128, 1000 - (expSize * 2), 0);
	} else if (weap->detonationType == DETONATIONTYPE_SLOWBURN) {
		//play_sample ((SAMPLE *) _global->SOUND[9].dat, 128 + (expSize / 2), 128, 1000 - (expSize * 2), 0);
	} else {
		play_sample ((SAMPLE *) _global->SOUND[WEAPONSOUNDS + sound].dat, 255, 128, 1000, 0);
	}
}

void MISSILE::draw (BITMAP *dest)
{
	if (!destroy) {
		if (y < MENU) {
			BITMAP *bbitmap = _bitmap;
			double by = y;
			int bangle = (int)angle;

			_bitmap = (BITMAP*)_global->gfxData.M[3].dat;
			y = (double)MENU + (_bitmap->h / 2);
			angle = 0;

			VIRTUAL_OBJECT::draw (dest);
			_bitmap = bbitmap;
			y = by;
			angle = (double)bangle;
		} else {

			if (weap->detonationType == DETONATIONTYPE_SLOWBURN) {
				circlefill (_env->db, (int)x, (int)y, expSize, makecol (255, 0, 0));
				setUpdateArea ((int)x - (expSize + 1), (int)y - (expSize + 1), (expSize + 1) * 2, (expSize + 1) * 2);
				requireUpdate ();
			} else {
				VIRTUAL_OBJECT::draw (dest);
			}
		}
	}
}

int MISSILE::isSubClass (int classNum)
{
	if (classNum == MISSILE_CLASS)
		return (TRUE);
	else
		return (FALSE);
		//return (PHYSICAL_OBJECT::isSubClass (classNum));
}
